@@ -30,11 +30,15 @@ pub enum Cmd<'a> {
     StartTls,
     Quit,
     Vrfy,
+    AuthLogin {
+        username: String,
+    },
     AuthPlain {
         authorization_id: String,
         authentication_id: String,
         password: String,
     },
+    AuthLoginEmpty,
     AuthPlainEmpty,
     // Dummy command containing client authentication
     AuthResponse {
@@ -81,6 +85,7 @@ pub struct Session<H: Handler> {
 pub struct SessionBuilder {
     name: String,
     start_tls_extension: bool,
+    insecure_allow_plaintext_auth: bool,
     auth_mechanisms: Vec<AuthMechanism>,
 }
 
@@ -90,6 +95,7 @@ impl SessionBuilder {
         Self {
             name: name.into(),
             start_tls_extension: false,
+            insecure_allow_plaintext_auth: false,
             auth_mechanisms: Vec::with_capacity(4),
         }
     }
@@ -106,6 +112,21 @@ impl SessionBuilder {
         self
     }
 
+    /// Allow authentication over plaintext and advertise authentication mechanisms before a connection
+    /// was upgraded to TLS with STARTTLS.
+    ///
+    /// This allows supporting non-compliant clients that either don't attempt STARTTLS if their
+    /// preferred authentication mechanism is not advertised before using STARTTLS or that don't
+    /// support TLS at all.
+    ///
+    /// Don't use this option if you don't know what you are doing and why it is fine to do in your
+    /// case, because this opens up the potential for clients to accidentally send credentials over
+    /// an insecure connection even if they actually support TLS.
+    pub fn insecure_enable_plaintext_auth(&mut self) -> &mut Self {
+        self.insecure_allow_plaintext_auth = true;
+        self
+    }
+
     /// Build a new session to handle a connection from the given ip address
     pub fn build<H: Handler>(&self, remote: IpAddr, handler: H) -> Session<H> {
         Session {
@@ -115,6 +136,7 @@ impl SessionBuilder {
                 remote,
                 self.auth_mechanisms.clone(),
                 self.start_tls_extension,
+                self.insecure_allow_plaintext_auth,
             ),
         }
     }
@@ -379,12 +401,21 @@ mod tests {
                 INVALID_CREDENTIALS
             )
         }
+
+        fn auth_login(&mut self, username: &str, password: &str) -> Response {
+            ternary!(
+                username == "test" && password == "1234",
+                AUTH_OK,
+                INVALID_CREDENTIALS
+            )
+        }
     }
 
     fn new_auth_session(with_start_tls: bool) -> Session<AuthHandler> {
         let addr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1));
         let mut builder = SessionBuilder::new("some.domain");
         builder.enable_auth(AuthMechanism::Plain);
+        builder.enable_auth(AuthMechanism::Login);
         if with_start_tls {
             builder.enable_start_tls();
         }
@@ -409,6 +440,20 @@ mod tests {
         assert_state!(session.fsm.current_state(), SmtpState::HelloAuth);
     }
 
+    #[test]
+    fn auth_ehlo() {
+        let mut session = new_auth_session(true);
+        start_tls(&mut session);
+        let res = session.process(b"ehlo a.domain\r\n");
+        assert_eq!(res.code, 250);
+        assert_state!(session.fsm.current_state(), SmtpState::HelloAuth);
+        let greeting = String::from_utf8(res.buffer().unwrap()).unwrap();
+        assert_eq!(
+            greeting,
+            "250-server offers extensions:\r\n250-8BITMIME\r\n250 AUTH PLAIN LOGIN\r\n".to_string()
+        )
+    }
+
     #[test]
     fn auth_plain_param() {
         let mut session = new_auth_session(true);
@@ -421,6 +466,21 @@ mod tests {
         assert_state!(session.fsm.current_state(), SmtpState::Hello);
     }
 
+    #[test]
+    fn auth_login_param() {
+        let mut session = new_auth_session(true);
+        start_tls(&mut session);
+        let mut res = session.process(b"ehlo a.domain\r\n");
+        assert_eq!(res.code, 250);
+        assert_state!(session.fsm.current_state(), SmtpState::HelloAuth);
+        res = session.process(b"auth login dGVzdA==\r\n"); // "test"
+        assert_eq!(res, PASSWORD_AUTH_CHALLENGE);
+        assert_state!(session.fsm.current_state(), SmtpState::Auth);
+        res = session.process(b"MTIzNA==\r\n"); // "1234"
+        assert_eq!(res.code, 235);
+        assert_state!(session.fsm.current_state(), SmtpState::Hello);
+    }
+
     #[test]
     fn bad_auth_plain_param() {
         let mut session = new_auth_session(true);
@@ -433,6 +493,21 @@ mod tests {
         assert_state!(session.fsm.current_state(), SmtpState::HelloAuth);
     }
 
+    #[test]
+    fn bad_auth_login_param() {
+        let mut session = new_auth_session(true);
+        start_tls(&mut session);
+        let mut res = session.process(b"ehlo a.domain\r\n");
+        assert_eq!(res.code, 250);
+        assert_state!(session.fsm.current_state(), SmtpState::HelloAuth);
+        res = session.process(b"auth login dGVzdA==\r\n"); // "test"
+        assert_eq!(res, PASSWORD_AUTH_CHALLENGE);
+        assert_state!(session.fsm.current_state(), SmtpState::Auth);
+        res = session.process(b"YmFkLXBhc3N3b3Jk\r\n"); // "bad-password"
+        assert_eq!(res.code, 535);
+        assert_state!(session.fsm.current_state(), SmtpState::HelloAuth);
+    }
+
     #[test]
     fn auth_plain_challenge() {
         let mut session = new_auth_session(true);
@@ -451,6 +526,24 @@ mod tests {
         assert_state!(session.fsm.current_state(), SmtpState::Hello);
     }
 
+    #[test]
+    fn auth_login_challenge() {
+        let mut session = new_auth_session(true);
+        start_tls(&mut session);
+        let res = session.process(b"ehlo a.domain\r\n");
+        assert_eq!(res.code, 250);
+        assert_state!(session.fsm.current_state(), SmtpState::HelloAuth);
+        let res = session.process(b"auth login\r\n");
+        assert_eq!(res, USERNAME_AUTH_CHALLENGE);
+        assert_state!(session.fsm.current_state(), SmtpState::Auth);
+        let res = session.process(b"dGVzdA==\r\n"); // "test"
+        assert_eq!(res, PASSWORD_AUTH_CHALLENGE);
+        assert_state!(session.fsm.current_state(), SmtpState::Auth);
+        let res = session.process(b"MTIzNA==\r\n"); // "1234"
+        assert_eq!(res.code, 235);
+        assert_state!(session.fsm.current_state(), SmtpState::Hello);
+    }
+
     #[test]
     fn auth_without_tls() {
         let mut session = new_auth_session(true);
@@ -461,6 +554,34 @@ mod tests {
         assert_eq!(res.code, 503);
     }
 
+    #[test]
+    fn auth_insecure_without_tls() {
+        let addr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1));
+        let mut builder = SessionBuilder::new("some.domain");
+        builder.enable_auth(AuthMechanism::Plain);
+        builder.insecure_enable_plaintext_auth();
+        let mut session = builder.build(addr, AuthHandler {});
+        let mut res = session.process(b"ehlo a.domain\r\n");
+        assert_eq!(res.code, 250);
+        assert_state!(session.fsm.current_state(), SmtpState::HelloAuth);
+        res = session.process(b"auth plain dGVzdAB0ZXN0ADEyMzQ=\r\n");
+        assert_eq!(res.code, 235);
+        assert_state!(session.fsm.current_state(), SmtpState::Hello);
+    }
+
+    #[test]
+    fn auth_insecure_without_auth() {
+        let addr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1));
+        let mut builder = SessionBuilder::new("some.domain");
+        builder.insecure_enable_plaintext_auth();
+        let mut session = builder.build(addr, AuthHandler {});
+        let mut res = session.process(b"ehlo a.domain\r\n");
+        assert_eq!(res.code, 250);
+        assert_state!(session.fsm.current_state(), SmtpState::Hello);
+        res = session.process(b"auth plain dGVzdAB0ZXN0ADEyMzQ=\r\n");
+        assert_eq!(res.code, 503);
+    }
+
     #[test]
     fn bad_auth_plain_challenge() {
         let mut session = new_auth_session(true);
@@ -472,6 +593,38 @@ mod tests {
         assert_state!(session.fsm.current_state(), SmtpState::HelloAuth);
     }
 
+    #[test]
+    fn bad_auth_login_username_challenge() {
+        let mut session = new_auth_session(true);
+        start_tls(&mut session);
+        session.process(b"ehlo a.domain\r\n");
+        let res = session.process(b"auth login\r\n");
+        assert_eq!(res, USERNAME_AUTH_CHALLENGE);
+        assert_state!(session.fsm.current_state(), SmtpState::Auth);
+        let res = session.process(b"YmFkLXVzZXJuYW1l\r\n"); // "bad-username"
+        assert_eq!(res, PASSWORD_AUTH_CHALLENGE);
+        assert_state!(session.fsm.current_state(), SmtpState::Auth);
+        let res = session.process(b"MTIzNA==\r\n"); // "1234"
+        assert_eq!(res.code, 535);
+        assert_state!(session.fsm.current_state(), SmtpState::HelloAuth);
+    }
+
+    #[test]
+    fn bad_auth_login_password_challenge() {
+        let mut session = new_auth_session(true);
+        start_tls(&mut session);
+        session.process(b"ehlo a.domain\r\n");
+        let res = session.process(b"auth login\r\n");
+        assert_eq!(res, USERNAME_AUTH_CHALLENGE);
+        assert_state!(session.fsm.current_state(), SmtpState::Auth);
+        let res = session.process(b"dGVzdA==\r\n"); // "test"
+        assert_eq!(res, PASSWORD_AUTH_CHALLENGE);
+        assert_state!(session.fsm.current_state(), SmtpState::Auth);
+        let res = session.process(b"YmFkLXBhc3N3b3Jk\r\n"); // "bad-password"
+        assert_eq!(res.code, 535);
+        assert_state!(session.fsm.current_state(), SmtpState::HelloAuth);
+    }
+
     #[test]
     fn rset_with_auth() {
         let mut session = new_auth_session(true);
@@ -487,4 +640,3 @@ mod tests {
         assert_state!(session.fsm.current_state(), SmtpState::HelloAuth);
     }
 }
-
